/* findFile - search set of paths for a file
 * Copyright (C) 2013-2020 FARGOS Development, LLC
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <iostream>
#ifndef _WIN32
#include <unistd.h>
#endif


static const char srcID[] = "$Id: findFile.cpp fedb7e888f79bb34c9886e007f7de64396da79f7 Sat Jun 1 00:32:15 2019 -0400 <gcc@fargos.net> $";

int elementCount;
char assembleBuffer[65535];
char *pathEntry[1024];
int bufferLen = 0;

static bool pathExists(const char *fileName)
{
	int rc = access(fileName, F_OK);
	if (rc == 0) return (true);
	return (false);
}

static const char *findPathElement(const char *element)
{
	for(int i=0;i<elementCount;i+=1) {
		if (strcmp(element, pathEntry[i]) == 0) { // found existing
			return (pathEntry[i]);
		}
	}
	return (0); // not found
}

static void addPathElem(const char *element)
{
	const char *prior;
	const char *cp = element;
	pathEntry[elementCount] = assembleBuffer + bufferLen;
	int startOffset = bufferLen;
	while (*cp != '\0') {
		if (*cp == ':') { // end of element
			assembleBuffer[bufferLen++] = '\0';
			prior = findPathElement(pathEntry[elementCount]);
			if (prior == 0) { // keep this one
				elementCount += 1;
				pathEntry[elementCount] = assembleBuffer + bufferLen;
				startOffset = bufferLen;
			} else { // discard
				bufferLen = startOffset;
			}
			cp += 1;
			continue;
		}
		assembleBuffer[bufferLen++] = *(cp++);
	}
	assembleBuffer[bufferLen++] = '\0';
	prior = findPathElement(pathEntry[elementCount]);
	if (prior == 0) { // keep this one
		elementCount += 1;
	}
}

int main(int argc, const char *argv[])
{
#ifdef _WIN32
    _setmode(_fileno(stdout), _O_BINARY);
#endif
	char pathSeparator = ':';
	int reportLimit = -1;
	bool sawSuffix = false;
	const char *suffix = 0;

	for(int i=1;i<argc;i+=1) {
		if ((strcmp(argv[i], "-h") == 0) ||
			(strcmp(argv[i], "-help") == 0) ||
			(strcmp(argv[i], "--help") == 0)) {
			fprintf(stderr, "usage: %s [-1 | --all ] [-s separator] fileOrDirName path [...]\n", argv[0]);
			fprintf(stderr, "  path can contain contiguous elements, default separator is a colon\n");
			fprintf(stderr, "  -s separator = define alternate path separator character\n");
			fprintf(stderr, "  -1 = report first entry found only -2 = report first two, etc.)\n");
			fprintf(stderr, "  --all = report all entries found (default)\n");
			fprintf(stderr, "  NOTE:  implemented via native executable\n");
			fprintf(stderr, "  ID=\"%s\"\n", srcID);
			return (1);
		}
		if (strcmp(argv[i], "-s") == 0) {
			pathSeparator = argv[i+1][0];
			i += 1;
			continue;
		}
		if ((strcmp(argv[i], "-all") == 0) ||
			(strcmp(argv[i], "--all") == 0)) {
			reportLimit = -1;
			continue;
		}
		if (argv[i][0] == '-') {
			if ((argv[i][1] >= '0') && (argv[i][1] <= '9')) {
				reportLimit = strtol(argv[i]+1, 0, 0);
				continue;
			}
			fprintf(stderr, "%s: unrecognized option %s\n", argv[0], argv[i]);
			return (1);
		}
		if (sawSuffix == false) {
			if (argv[i][0] == '/') { // absolute pathname
				if (pathExists(argv[i])) {
					std::cout << argv[i] << "\n";
					std::cout.flush();
					return (0);
				}
			}
			suffix=argv[i];
			sawSuffix = true;
		} else {
			addPathElem(argv[i]);
		}
	}
	if (suffix == 0) {
		fprintf(stderr, "%s: no target file specified.\n", argv[0]);
		return (1);
	}
	if (elementCount == 0) {
		fprintf(stderr, "%s: no path elements specified.\n", argv[0]);
		return (1);
	}
	int foundCount = 0;
	int exitResult = 1;
	for(int i=0;i<elementCount;i+=1) {
		char bfr[2048];
		snprintf(bfr, sizeof(bfr), "%s/%s", pathEntry[i], suffix);
		if (pathExists(bfr)) {
			std::cout << bfr << "\n";
			exitResult = 0; // found something
			foundCount += 1;
			if (foundCount == reportLimit) { // all done
				break;	// all done
			}
		}
	}
	std::cout.flush();
	return (exitResult);
}
