/* pathCondense - condense path set to simplest representation.
 * Removes duplicate entries and can delete references to elements that
 * do not exist.
 * Copyright (C) 2013 - 2020 FARGOS Development, LLC
 */

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <iostream>
#ifndef _WIN32
#include <unistd.h>
#endif

static const char srcID[] = "$Id: pathCondense.cpp 6a231b4e75d5348ade7f3a7369236ec791a56f82 Sun Jul 26 11:11:25 2020 -0400 <gcc@fargos.net> $";

#define MAX_POSSIBLE_ELEMENTS 1024
#define MAX_PATH_LEN 65535

static int elementCount;
static char assembleBuffer[MAX_PATH_LEN + 1];
static char *pathEntry[MAX_POSSIBLE_ELEMENTS];
static int bufferLen = 0;
static bool mustExist = false;
static int verboseLevel = 0;
static const char *progName = nullptr;

static const char *findPathElement(const char *element)
{
    for(int i=0;i<elementCount;i+=1) {
        if (strcmp(element, pathEntry[i]) == 0) { // found existing
            return (pathEntry[i]);
        }
    }
    return (nullptr); // not found
}

int testFileExistence(const char *fileName)
{
#if defined(_WIN32) && (__CYGWIN__ == 0)
    if ((strncmp(fileName, "/bin", 4) == 0) ||
        (strncmp(fileName, "/sbin", 5) == 0) ||
        (strncmp(fileName, "/home", 5) == 0) ||
        (strncmp(fileName, "/usr", 4) == 0)) {
        return (0); // claim always present
    }
    if (strncmp(fileName, "/cygdrive/", 10) != 0) {
#endif
    int rc = access(fileName, F_OK);
    return (rc);
#if defined(_WIN32) && (__CYGWIN__ == 0)
    }
    // begins with cygdrive, convert /cygdrive/X/ to X:/
    char fileWithDrivePrefix[8192];
    memccpy(fileWithDrivePrefix + 2, fileName + 11, 0, sizeof(fileWithDrivePrefix) - 1);
    fileWithDrivePrefix[0] = fileName[10];
    fileWithDrivePrefix[1] = ':';
    fileWithDrivePrefix[sizeof(fileWithDrivePrefix) - 1] = '\0';
    int rc2 = access(fileWithDrivePrefix, F_OK);
//fprintf(stderr, "fileWithDrivePrefix=%s rc=%d\n", fileWithDrivePrefix, rc2);
    return (rc2);
#endif
}

static void addPathElem(const char *element, const char inputSeparator=':')
{
    const char *prior;
    const char *cp = element;
    pathEntry[elementCount] = assembleBuffer + bufferLen;
    int startOffset = bufferLen;
    while ((*cp != '\0') && (bufferLen < MAX_PATH_LEN)) {
        if (*cp == inputSeparator) { // end of element
            assembleBuffer[bufferLen++] = '\0';
            if (verboseLevel > 0) {
                fprintf(stderr,"%s\n", pathEntry[elementCount]);
                fflush(stderr);
            }
            prior = findPathElement(pathEntry[elementCount]);
            if (prior == nullptr) { // keep this one
                int rc = 0;
                if (mustExist == true) {
                    rc = testFileExistence(pathEntry[elementCount]);
                }
                if (rc == 0) { // file exists
                    if (elementCount < MAX_POSSIBLE_ELEMENTS) {
                        elementCount += 1;
                    } else {
                        fprintf(stderr,"%s: too many path elements\n", progName);
                    }
                }
                pathEntry[elementCount] = assembleBuffer + bufferLen;
                startOffset = bufferLen;
            } else { // discard
                bufferLen = startOffset;
            }
            cp += 1;
            continue;
        }
        assembleBuffer[bufferLen++] = *(cp++);
    }
    assembleBuffer[bufferLen++] = '\0';
    if (bufferLen >= MAX_PATH_LEN) {
        fprintf(stderr, "%s: assembled path too long\n", progName);
        return;
    }
    if (verboseLevel > 0) {
        fprintf(stderr,"%s\n", pathEntry[elementCount]);
        fflush(stderr);
    }
    prior = findPathElement(pathEntry[elementCount]);
    if (prior == nullptr) { // keep this one
        int rc = 0;
        if (mustExist == true) {
            rc = testFileExistence(pathEntry[elementCount]);
        }
        if (rc == 0) { // file exists
            elementCount += 1;
        }
    }
}

int main(int argc, const char *argv[])
{
    progName = argv[0];
    char pathSeparator = ':';
    char inputSeparator = ':';

#ifdef _WIN32
    _setmode(_fileno(stdout), _O_BINARY);
#endif
    for(int i=1;i<argc;i+=1) {
        if ((strcmp(argv[i], "-h") == 0) ||
            (strcmp(argv[i], "-help") == 0) ||
            (strcmp(argv[i], "--help") == 0)) {
            fprintf(stderr, "usage: %s [{{-s separator} | --space}] [-x] dir [...]\n", argv[0]);

            fprintf(stderr, "  -s = specifies output separator for path elements, defaults to \"%c\"\n", pathSeparator);
            fprintf(stderr, "  --is = specifies input separator for path elements, defaults to \"%c\"\n", inputSeparator);
            fprintf(stderr, "  --space = use space as output separator\n");
            fprintf(stderr, "  -x - only include filesystem elements that exist\n");
            fprintf(stderr, "  NOTE:  implemented via native executable\n");
            fprintf(stderr, "  ID=\"%s\"\n", srcID);
            return (1);
        }   
        if ((strcmp(argv[i], "-s") == 0) ||
           (strcmp(argv[i], "-os") == 0) ||
           (strcmp(argv[i], "--os") == 0)) {
            pathSeparator = argv[i+1][0];
            i += 1;
            continue;
        }
        if ((strcmp(argv[i], "-is") == 0) ||
           (strcmp(argv[i], "--is") == 0)) {
            inputSeparator = argv[i+1][0];
            i += 1;
            continue;
        }
        if (strcmp(argv[i], "-x") == 0) {
            mustExist = true;
            continue;
        }
        if (strcmp(argv[i], "--verbose") == 0) {
            verboseLevel += 1;
            continue;
        }
        if ((strcmp(argv[i], "-space") == 0) ||
            (strcmp(argv[i], "--space") == 0)) {
            pathSeparator = ' ';
            continue;
        }
        if (argv[i][0] == '-') {
            fprintf(stderr, "%s: unrecognized option: \"%s\"\n",
                    argv[0], argv[i]);
            return (1);
        }
        addPathElem(argv[i], inputSeparator);
    } // end for each command line argument

    for(int i=0;i<elementCount;i+=1) {
        if (i != 0) {
            std::cout << pathSeparator;
        }
        std::cout << pathEntry[i];
    }
    std::cout << "\n";
    std::cout.flush();
    return (0);
}
/* vim: set expandtab shiftwidth=4 tabstop=4: */
